'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _utils = require('./utils');

var _sideEffect = require('./sideEffect');

/**
 * Iterator mapper, maps iterator to value
 *
 * @param {Object} Iterator
 * @returns {Any} Iterator Value
 */
var mapValue = function mapValue(iterator) {
  return iterator.value;
};

/**
 * Reducer enhancer. Iterates over generator like reducer reduction and accumulates
 * all the side effects and reduced application state.
 *
 * @param {Function} Root reducer in form of generator function
 * @param {Function} Function used for deferring effects accumulated from Reduction
 * @returns {Function} Reducer
 */

exports.default = function (rootReducer, deferEffects) {
  return function (appState, action) {
    (0, _utils.invariant)((0, _utils.isFunction)(rootReducer), 'Provided root reducer is not a function.');

    // Calling the Root reducer should return an Iterable
    var reduction = rootReducer(appState, action);

    if ((0, _utils.isIterable)(reduction)) {
      // Iterable returned by Root reducer is mapped into array of values.
      // Last value in the array is reduced application state all the other values
      // are just side effects.
      var effects = (0, _utils.mapIterable)(reduction, mapValue);
      var newState = effects.pop();

      (0, _utils.invariant)(effects.every(_sideEffect.isSideEffect), 'Yielded side effects must always be created using sideEffect function');

      deferEffects(effects);

      (0, _utils.invariant)(!(0, _utils.isUndefined)(newState), 'Root reducer does not return new application state. Undefined is returned');

      return newState;
    } else {
      console.warn('createEffectCapableStore enhancer from redux-side-effects package is used, ' + 'yet the provided root reducer is not a generator function');

      return reduction;
    }
  };
};